/**** << Text Formating & Printing Utilities >> ****
 *
 *  psconv.c	Ver 2.23
 *                                                  Oct 28 1991
 *                                      mod:        May 11 1994
 ************************************************** tonooka ***********/
/*
 *	Copyright (c) 1994 Yasuhiro Tonooka (tonooka@msi.co.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#if !defined lint
static char *copyright = "@(#)Copyright (c) 1994 Yasuhiro Tonooka";
static char *sccsid = "@(#)psconv.c 1.7 (Y.Tonooka) 5/11/94";
#endif

/*-
 * psconv recognizes four types of input characters: `regular',
 * `underlined', `boldfaced', and `both underlined and boldfaced',
 * which are implemented by the `overstriking' mechanism.  For example,
 * 
 *	`_^HX', `X^H_', `XX^H^H__', `[]^H^H__', `XXXX^M____'
 * 
 * are all regarded as underlined characters,
 * 
 *	`X^HX', `XX^H^HXX', `[]^H^H[]', `XXXX^MXXXX'
 * 
 * are boldfaced, and
 * 
 *	`_^HX^HX', `__^H^H[]^H^H[]', `____^MXXXX^MXXXX'
 * 
 * are both underlined and boldfaced.  Here, X stands for an arbitrary
 * ASCII printing character except space, [] for an arbitrary em-size
 * kanji character, ^H for a single BACKSPACE, and ^M for
 * CARRAGERETURN.
 * 
 * To underline em-size kanji characters, two underscores are encouraged
 * to be used, although "psconv" is designed to accept single
 * underscore sequences like:
 * 
 *	`_^H[]' or `[]^H_'
 * 
 * also.
 * 
 * The mapping of input character types to output font types can be
 * reassigned using "-f" option.
 * 
 * By default, the `regular' type is mapped to `Courier', the
 * `underlined' type to `Courier-Oblique', the `boldfaced' to
 * `Courier-Bold', and the `combined' type to `Courier-BoldOblique',
 * all without underlining.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>

typedef int bool;

#define putstr(s)	fputs(s, stdout)

/*-
 *  OPTIONS:
 *	-Vxx-yy
 *	    or -Vxx-yy-cc
 *		`xx * yy' or `xx * yy * cc' portrait format.
 *	-Hxx-yy
 *	    or -Hxx-yy-cc
 *		`xx * yy' or `xx * yy * cc' landscape format.
 *		Followings are the options for user's convenience:
 *		-V	80  * 66	portrait (default)
 *		-Vw	132 * 100	portrait wide
 *		-V2	80  * 132 * 2	portrait two columns
 *		-V3	80  * 200 * 3	portrait three columns
 *		-H	132 * 66	landscape wide
 *		-H2	80  * 66  * 2	landscape two columns
 *		-H3	80  * 100 * 3	landscape three columns
 *		-H4	80  * 132 * 4	landscape four columns
 *
 *	-s	Place separator (dash) between adjacent columns.
 *
 *	-n	Eliminate margins on four sides of a page.
 *
 *	-h	Hiragana is used for "hankaku kana" characters.
 *
 *	-j	Do not modify the shape of kanji characters:  let them
 *		be square.
 *
 *	-`num'	Expand TAB characters at every `num'-th character
 *		position.  Defaults to 8.
 *
 *	-mxx	Place margin of xx characters between adjacent
 *		columns instead of 8 (default).
 *
 *	-fxxx	Reassign mapping of input character types to output
 *		font types.  This option goes like:
 *		    -fB:BU:IU:R
 *		Each `:'-separated field in order stands for input of:
 *		    regular			(x)
 *		    underlined			(_^Hx)
 *		    boldfaced			(x^Hx)
 *		    boldfaced & underlined	(_^Hx^Hx)
 *		Font type characters are:
 *		    R	regular
 *		    B	boldfaced (Bold)
 *		    I	italic (Oblique)
 *		    U	underlined
 *		`B', `I' and `U' can be combined together.  If a type
 *		specification is omitted for a field, the default
 *		mapping
 *		    -fR:I:B:BI
 *		is understood for that field.  `R' alone must be used
 *		explicitly to specify regular courier.
 *		    -fB
 *		is the same as
 *		    -fB:::
 */

/**********************************************************************
 *                                                                    *
 *  Global Variables                                                  *
 *                                                                    *
 **********************************************************************/
char *progname;

enum orient {
    PORTRAIT,
    LANDSCAPE,
} orient = PORTRAIT;
int width = 80, numrow = 66;
int column = 1;			/* number of columns */
int tabstop = 8;
bool nomarg = 0;		/* margin switch */
int margin = 8;			/* margin between columns */
bool dash = 0;			/* whether to put lines in between */
bool square = 0;		/* make kanji square */

bool intmask = 0;		/* masked when SIGINT is prohibited */
bool intflag = 0;		/* set when SIGINT is received */

int totalcount = 0;		/* total page count */

extern unsigned short katakana[], hiragana[];
				/* JIS X0201 kana to EUC kana table */
unsigned short *kanatable = katakana;	/* kana table to be used */

/*
 * Character types:
 */
enum chartype {
    ASCII, KANJI, HANKAKU
};
/*
 * Font types:
 */
#define I	1		/* italic */
#define B	2		/* bold */
#define U	4		/* underline */

/*
 * Table of font names:
 */
char *fonttable[3][8] = {
    "R N ",  "I N ",  "B N ",  "C N ",
    "R U ",  "I U ",  "B U ",  "C U ",
    "KR N ", "KI N ", "KB N ", "KC N ",
    "KR U ", "KI U ", "KB U ", "KC U ",
    "HR N ", "HI N ", "HB N ", "HC N ",
    "HR U ", "HI U ", "HB U ", "HC U ",
};
/*
 * Font map:
 */
int fontmap[] = {
    0,  I,  B,  I | B		/* default mapping */
};

void convert();
void fmtopt(), fontopt();
void sigint();
void header();
void prologue();
void trailer();
void reset();
bool ff();
bool cr();
void tab();
void startpage();
void startcolumn();
void outnonprint();
void outchar();
void outhankana();
void outkanji();

void interrupt();
void error();

void beginstr();
void endstr();
void initbuf();

/**********************************************************************
 *                                                                    *
 *  Main & Command Line Argument Handling Routines                    *
 *                                                                    *
 **********************************************************************/
/*-==(spec:\\:":":\(:\):\[:\]::::psconv)==-----------------------------
    NAME
	main
 ---------------------------------------------------------------------*/
int main(c, v)
    int c;
    register char *v[];
{
    register char *s;
    register FILE *iop;
    int ret;
    int dev, ino = -1;
    struct stat sbuf;
    void (*sigold)();

    progname = *v++;
    /*
     * Process command line options.
     */
    for (; --c; v++) {
	/*
	 * With -, input is taken from stdin like cat(1).
	 */
	if ((*v)[0] != '-' || (*v)[1] == '\0')
	    break;
	/*
	 * Set orientation to portrait and set the size
	 */
	if ((*v)[1] == 'V') {
	    orient = PORTRAIT;
	    if ((*v)[2] == '\0') {
		width = 80;
		numrow = 66;
		column = 1;
	    } else
		if (strcmp(*v + 2, "w") == 0) {
		    width = 132;
		    numrow = 100;
		    column = 1;
		} else if (strcmp(*v + 2, "2") == 0) {
		    width = 80;
		    numrow = 132;
		    column = 2;
		} else if (strcmp(*v + 2, "3") == 0) {
		    width = 80;
		    numrow = 200;
		    column = 3;
		} else
		    fmtopt(*v + 2);
	    continue;
	}
	/*
	 * Set orientation to landscape and set the size
	 */
	if ((*v)[1] == 'H') {
	    orient = LANDSCAPE;
	    if ((*v)[2] == '\0') {
		width = 132;
		numrow = 66;
		column = 1;
	    } else
		if (strcmp(*v + 2, "2") == 0) {
		    width = 80;
		    numrow = 66;
		    column = 2;
		} else if (strcmp(*v + 2, "3") == 0) {
		    width = 80;
		    numrow = 100;
		    column = 3;
		} else if (strcmp(*v + 2, "4") == 0) {
		    width = 80;
		    numrow = 132;
		    column = 4;
		} else
		    fmtopt(*v + 2);
	    continue;
	}
	/*
	 * Remapping fonts:
	 */
	if ((*v)[1] == 'f') {
	    fontopt(*v + 2);
	    continue;
	}
	/*
	 * -m option specifies margin.
	 */
	if ((*v)[1] == 'm') {
	    if (!isdigit((*v)[2]) ||
		    (margin = atoi(*v + 2)) < 0) {
		error("bad margin specification", *v + 2);
		exit(1);
	    }
	    continue;
	}
	/*
	 * TAB characters are expanded to SPACE characters.  Tabstops
	 * are assumed to be at every eighth character posision
	 * unless they are specified with option -`num'.
	 */
	if (isdigit((*v)[1])) {
	    if ((tabstop = atoi(*v + 1)) < 1) {
		error("bad tabstop specification", *v + 1);
		exit(1);
	    }
	    continue;
	}
	/*
	 * One letter options:
	 */
	for (s = *v + 1; *s; s++) {
	    switch (*s) {
	    case 'j':
		/*
		 * Make kanji characters square.
		 */
		square = 1;
		break;
	    case 'h':
		/*
		 * Hiragana instead of katakana for "hankaku kana".
		 */
		kanatable = hiragana;
		break;
	    case 'n':
		/*
		 * -n option disables taking margin.
		 */
		nomarg = 1;
		break;
	    case 's':
		/*
		 * Draw separator (dash) in between columns.
		 */
		dash = 1;
		break;
	    default:
		s[1] = '\0';
		error("bad option", s);
		exit(1);
	    }
	}
    }
    /*
     * Get some info on output file.
     */
    if (fstat(fileno(stdout), &sbuf) == 0) {
	sbuf.st_mode &= S_IFMT;
	if (sbuf.st_mode != S_IFCHR && sbuf.st_mode != S_IFBLK) {
	    dev = sbuf.st_dev;
	    ino = sbuf.st_ino;
	}
    }
    /*
     * Main loop.
     */
    sigold = (void (*)()) signal(SIGINT, sigint);
    intmask = 1;
    header();
    prologue();
    intmask = 0;
    if (intflag)
	interrupt();
    ret = 0;
    do {
	if (c == 0 || strcmp(*v, "-") == 0) {
	    /*
	     * Stdin:  If at EOF and tty, clear EOF.
	     */
	    if (feof(stdin))
		if (isatty(fileno(stdin)))
		    clearerr(stdin);
		else
		    continue;
	    iop = stdin;
	} else
	    /*
	     * Open a file.
	     */
	    if ((iop = fopen(*v, "r")) == NULL) {
		perror(*v);
		ret = 1;
		continue;
	    }
	if (fstat(fileno(iop), &sbuf) == 0) {
	    /*
	     * Get some info on input file, and see if it is a
	     * directory. 
	     */
	    if ((sbuf.st_mode & S_IFMT) == S_IFDIR) {
		fprintf(stderr,
			"%s: read error on %s: Is a directory\n",
			progname, c ? *v : "standard input");
		if (iop != stdin)
		    fclose(iop);
		ret = 1;
		continue;
	    }
	    /*
	     * Compare the info of input with that of output, and see
	     * if they are identical. 
	     */
	    if ((sbuf.st_mode & S_IFMT) == S_IFREG &&
		    sbuf.st_dev == dev && sbuf.st_ino == ino) {
		fprintf(stderr, "%s: input %s is output\n", progname,
			c ? *v : "-");
		if (iop != stdin)
		    fclose(iop);
		ret = 1;
		continue;
	    }
	}
	/*
	 * Do the job here!
	 */
	convert(iop);
	if (iop != stdin)
	    fclose(iop);
    } while (v++, --c > 0);
    intmask = 1;
    trailer();
    (void) signal(SIGINT, sigold);
    intmask = 0;
    if (intflag)
	interrupt();
    if (ferror(stdout)) {
	error("output write error", (char *) NULL);
	ret = 1;
    }
    return (ret);
}

/*-==(spec:\\:":":\(:\):\[:\]::::psconv)==-----------------------------
    NAME
	fmtopt - handle format option
 ---------------------------------------------------------------------*/
void fmtopt(str)
    char *str;
{
    int i;
    register char *s;

    s = str;
    if ((i = atoi(s)) <= 0) {
	error("bad size specification", str);
	exit(1);
    }
    width = i;
    while (isdigit(*s))
	s++;
    if (*s++ != '-' || (i = atoi(s)) <= 0) {
	error("bad size specification", str);
	exit(1);
    }
    numrow = i;
    while (isdigit(*s))
	s++;
    if (*s++ == '-') {
	if ((i = atoi(s)) <= 0) {
	    error("bad size specification", str);
	    exit(1);
	}
	column = i;
    } else
	column = 1;
}

/*-==(spec:\\:":":\(:\):\[:\]::::psconv)==-----------------------------
    NAME
	fontopt - handle font remapping option
 ---------------------------------------------------------------------*/
void fontopt(str)
    char *str;
{
    register char *s;
    register int type;
    register bool b;
    register int i;

    for (s = str, i = 0; *s && i < 4; i++) {
	type = 0;
	b = 0;
	for (; *s && *s != ':'; s++) {
	    switch (*s) {
	    case 'I':		/* italic */
		type |= I;
		break;
	    case 'B':		/* boldface */
		type |= B;
		break;
	    case 'U':		/* underlined */
		type |= U;
		break;
	    case 'R':		/* regular */
		if (!b)
		    break;
		/* fall thru */
	    default:
		error("bad font specification", str);
		exit(1);
	    }
	    b = 1;
	}
	if (b)
	    fontmap[i] = type;
	if (*s)
	    s++;
    }
}

/*-==(spec:\\:":":\(:\):\[:\]::::psconv)==-----------------------------
    NAME
	error
 ---------------------------------------------------------------------*/
void error(expln, stuff)
    char *expln, *stuff;
{
    if (stuff)
	fprintf(stderr, "%s: %s: %s\n", progname, stuff, expln);
    else
	fprintf(stderr, "%s: %s\n", progname, expln);
}

/**********************************************************************
 *                                                                    *
 *  Filter & Input Routines                                           *
 *                                                                    *
 **********************************************************************/
#define LENBUF	BUFSIZ
unsigned short cbuf[LENBUF];	/* process code buffer */
unsigned char abuf[LENBUF];	/* attribute buffer */
#define ULINE	1
#define BOLD	2

#define EMEND	0xffff		/* second column of em letter */
#define SS2	0x8e
#define SS3	0x8f

bool inmode;			/* whether we're in a string or not */
enum chartype currmode;		/* ASCII, HANKAKU, KANJI */
int currfont;			/* 0 to 7 */

int charcount, linecount, columncount, pagecount;

#if defined iskanji
#undef iskanji
#endif
#define iskanji(c)	((c) >= 0x100)
#define ishankana(c)	((c) > 0xa0 && (c) < 0xe0)

/*-==(spec:\\:":":\(:\):\[:\]::::psconv)==-----------------------------
    NAME
	convert
 ---------------------------------------------------------------------*/
void convert(fp)
    FILE *fp;
{
    register int c;
    register int i;
    int font;
    int end;
    bool doff, dostart, done;

    inmode = 0;
    pagecount = 0;
    currmode = ASCII;
    currfont = 0;
    initbuf(LENBUF);
    doff = 0;
    dostart = 1;
    done = 0;

    while ((end = getline(fp)) > 0) {
	for (i = 0; i < end; i++) {
	    intmask = 1;
	    /*
	     * Do `formfeed' only when we have more characters to
	     * print.
	     */
	    if (doff) {
		dostart = ff();
		doff = 0;
	    }
	    /*
	     * Do not start a page until at least one character is
	     * read, so nothing will be printed for empty input.
	     */
	    if (dostart) {
		startpage();
		done = 1;
		dostart = 0;
	    }
	    c = cbuf[i];
	    font = fontmap[abuf[i]];
	    if (iskanji(c)) {
		outkanji(c, font);
		i++;
	    } else if (isprint(c))
		outchar(c, font);
	    else if (c == '\f')
		doff = 1;
	    else if (c == '\n')
		doff = cr();
	    else if (ishankana(c))
		outhankana(c, font);
	    else
		outnonprint(c, font);
	    intmask = 0;
	    if (intflag)
		interrupt();
	}
	initbuf(end);
    }
    intmask = 1;
    if (done) {
	if (inmode)
	    endstr();
	putstr("grestore\n");
	putstr("showpage\n");
	putstr("restore\n");
    }
    intmask = 0;
    if (intflag)
	interrupt();
}

/*-==(spec:\\:":":\(:\):\[:\]::::psconv)==-----------------------------
    NAME
	getline
    DESCRIPTION
	Reads characters onto cbuf[] upto NEWLINE, FORMFEED, escape
	sequence or EOF.
	An attribute of each character is set onto abuf[].
	No checking on buffer overflow.
 ---------------------------------------------------------------------*/
int getline(fp)
    FILE *fp;
{
    register int c;
    register int col = 0;
    int maxcol = 0;

    while ((c = getkc(fp)) != EOF) {
	switch (c) {
	case '\n':
	case '\f':
	    cbuf[maxcol++] = c;
	    goto ret;
	case '\b':
	    if (col > 0)
		--col;
	    continue;
	case '\r':
	    col = 0;
	    continue;
	case '\t':
	    do {
		if (cbuf[col] == '_') {
		    cbuf[col] = ' ';
		    abuf[col] |= ULINE;
		} else if (cbuf[col] == '\0')
		    cbuf[col] = ' ';
		col++;
	    } while (col % tabstop);
	    break;
	case ' ':
	    if (cbuf[col] == '_') {
		cbuf[col] = ' ';
		abuf[col] |= ULINE;
	    } else if (cbuf[col] == '\0')
		cbuf[col] = ' ';
	    col++;
	    break;
	case '_':
	    switch (cbuf[col]) {
	    case '\0':
		cbuf[col] = '_';
		break;
	    case '_':
		if (abuf[col] & (ULINE | BOLD))
		    abuf[col] |= ULINE | BOLD;
		else if (col && (cbuf[col - 1] != EMEND ?
			(abuf[col - 1] & (ULINE | BOLD)) == BOLD :
			(abuf[col - 2] & (ULINE | BOLD)) == BOLD))
		    abuf[col] |= BOLD;
		else
		    abuf[col] |= ULINE;
		break;
	    case EMEND:
		abuf[col - 1] |= ULINE;
		break;
	    default:
		abuf[col] |= ULINE;
		break;
	    }
	    col++;
	    break;
	default:
	    if (c < 0x100) {
		/*
		 * Single width character (ASCII or katakana)
		 */
		if (cbuf[col] == '_') {
		    cbuf[col] = c;
		    abuf[col] |= ULINE;
		} else if (cbuf[col] == c)
		    abuf[col] |= BOLD;
		else
		    cbuf[col] = c;
		col++;
		break;
	    }
	    /*
	     * Double width character (kanji)
	     */
	    if (cbuf[col] == EMEND)
		--col;
	    if (cbuf[col] == c)
		abuf[col] |= BOLD;
	    else {
		if (cbuf[col] == '_')
		    abuf[col] |= ULINE;
		cbuf[col] = c;
		cbuf[col + 1] = EMEND;
	    }
	    col += 2;
	    break;
	}
	if (col > maxcol)
	    maxcol = col;
    }
ret:
    cbuf[maxcol] = '\0';
    return (maxcol);
}

/*-==(spec:\\:":":\(:\):\[:\]::::psconv)==-----------------------------
    NAME
	getkc
    DESCRIPTION
	Gets one process code like JLE's getwc().  For EUC code set 3
	characters, 0xa2ae is always returned.
 ---------------------------------------------------------------------*/
int getkc(fp)
    register FILE *fp;
{
    register int c;

    if ((c = getc(fp)) == EOF || (c & 0x80) == 0)
	return (c);		/* ASCII */
    if (c == SS2)
	return (getc(fp));	/* en-katakana */
    if (c == SS3)		/* For SS3, '0xa2ae' is returned. */
	return (getc(fp) == EOF || getc(fp) == EOF ? EOF : 0xa2ae);
    if (c < 0xa1 || c == 0xff)
	return (c);		/* control chars */
    return (c << 8 | getc(fp));	/* em-kanji */
}

/*-==(spec:\\:":":\(:\):\[:\]::::psconv)==-----------------------------
    NAME
	initbuf
    DESCRIPTION
	Initializes "cbuf[]" and "abuf[]".
 ---------------------------------------------------------------------*/
void initbuf(end)
    register int end;
{
    register int i;

    for (i = 0; i < end; i++) {
	cbuf[i] = 0;
	abuf[i] = 0;
    }
}

/**********************************************************************
 *                                                                    *
 *  Interrupt Handling Routines                                       *
 *                                                                    *
 **********************************************************************/
/*-==(spec:\\:":":\(:\):\[:\]::::psconv)==-----------------------------
    NAME
	siginit
    DESCRIPTION
	A SIGINT handler.
 ---------------------------------------------------------------------*/
void sigint()
{
    (void) signal(SIGINT, SIG_IGN);
    if (intmask) {
	intflag = 1;
	return;
    }
    interrupt();
}

/*-==(spec:\\:":":\(:\):\[:\]::::psconv)==-----------------------------
    NAME
	interrupt
    DESCRIPTION
	Reset and exit.
 ---------------------------------------------------------------------*/
void interrupt()
{
    if (inmode) {
	fprintf(stderr, "SIGINT received in mode.\n");
	endstr();
    }
    reset();
    fprintf(stderr, "SIGINT received.\n");
    exit(1);
}

/**********************************************************************
 *                                                                    *
 *  Output Routines                                                   *
 *                                                                    *
 **********************************************************************/
#define SCALEFONT	10.0	/* point (6.6 in / 80 char) */
#define HEIGHT		10.0	/* in. */	/* * 11/10 */
#define WIDTH		6.6	/* in. */	/* * 11/10 */

/*-==(spec:\\:":":\(:\):\[:\]::::psconv)==-----------------------------
    NAME
	header
    DESCRIPTION
	Prints out header comment.
 ---------------------------------------------------------------------*/
void header()
{
    long t;
    char *user;
    extern char *ctime();
    extern char *getlogname();
    extern char *header_str[];

    t = time(NULL);
    user = getlogname();
    putstr(header_str[0]);
    putstr(user ? user : "???");
    putstr(header_str[1]);
    putstr(ctime(&t));
    putstr(header_str[2]);
    putstr(user ? user : "???");
    putstr(header_str[3]);
    putstr(ctime(&t));
    putstr(header_str[4]);
}

/*-==(spec:\\:":":\(:\):\[:\]::::psconv)==-----------------------------
    NAME
	prologue
 ---------------------------------------------------------------------*/
void prologue()
{
    double point;
    double feed;
    double shrink;
    int w;
    double x, y, s;
    int r;
    extern char prolog_str[];

    reset();
    w = width * column + margin * (column - 1);
    if (orient == LANDSCAPE) {
	point = 80.0 / w * SCALEFONT * (HEIGHT / WIDTH);
	feed = WIDTH / numrow * 72;
    } else {
	point = 80.0 / w * SCALEFONT;
	feed = HEIGHT / numrow * 72;
    }
    /*
     * When linefeed is too small for the point size of letters:
     */
    if (feed < point) {
	shrink = feed / point;
	feed = point;
    } else
	shrink = 1.0;
    putchar('\n');
    /*
     * Parameters.
     */
    putstr("\
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n\
%                                                                      %\n\
%   Parameters                                                         %\n\
%                                                                      %\n\
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n");
    printf("/point %g def\n", point);
    printf("/linefeed %g def\n", feed);
    if (orient == LANDSCAPE) {		/* horizontal */
	r = 90;
	if (nomarg)
	    x = 10.0, y = -30.0, s = 1.13;
	else
	    x = 19.0, y = -52.0, s = 1.09;
    } else {			/* vertical */
	r = 0;
	if (nomarg)
	    x = 23.0, y = 825.0, s = 1.133;
	else
	    x = 53.0, y = 780.0, s = 1.0;
    }
    y -= point * 0.8;
    printf("/base-r %d def\n", r);
    printf("/base-x %g def\n", x);
    printf("/base-y %g def\n", y);
    printf("/base-scale-x %g def\n", s);
    printf("/base-scale-y %g def\n", s * shrink);
    /*
     * Kanji & hankaku size:
     */
    printf("/kanji-height %g def\n", square ? 1.1 : 1.0 / 1.2);
    printf("/kanji-width %g def\n", square ? 1.1 : 1.2);
    printf("/hankaku-height %g def\n", 1.0 / 1.2);
    printf("/hankaku-width %g def\n", 0.75);
    /*
     * Parameters for separator:
     */
    printf("/column-interval %d def\n", width + margin);
    printf("/dash-count %d def\n", dash ? column : 0);
    printf("/dash-init %g def\n", width + margin / 2.0);
    printf("/dash-len %g def\n", feed * numrow);
#if defined DEBUG
    putstr("/#copies 0 def\t\t\t% debugging mode");
#endif
    putchar('\n');
    putstr("\
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n\
%                                                                      %\n\
%   Procedures                                                         %\n\
%                                                                      %\n\
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n");
    putstr(prolog_str);
}

/*-==(spec:\\:":":\(:\):\[:\]::::psconv)==-----------------------------
    NAME
	startpage
    DESCRIPTION
	Start a new page
 ---------------------------------------------------------------------*/
void startpage()
{
    pagecount++;
    totalcount++;
    printf("\n%%%%Page: %d %d\n", pagecount, totalcount);
    putstr("save\n");
    putstr("gsave\n");
    putstr("PAGE-SETUP\n");
    startcolumn();
    columncount = 0;
    linecount = 0;
    charcount = 0;
}

/*-==(spec:\\:":":\(:\):\[:\]::::psconv)==-----------------------------
    NAME
	startcolumn
 ---------------------------------------------------------------------*/
void startcolumn()
{
    /*
     * Set left margin to "n"-th character position.
     */
    putstr("\n% starting new column\n");
    putstr("COLUMN-SETUP\n");
    putstr(fonttable[(int) currmode][currfont]);
    putstr("MOVE\n");
}

/*-==(spec:\\:":":\(:\):\[:\]::::psconv)==-----------------------------
    NAME
	trailer
 ---------------------------------------------------------------------*/
void trailer()
{
    putstr("\n%%Trailer\n");
    reset();
    printf("\n%%%%Pages: %d\n", totalcount);
}

/*-==(spec:\\:":":\(:\):\[:\]::::psconv)==-----------------------------
    NAME
	ff - formfeed
 ---------------------------------------------------------------------*/
bool ff()
{
    if (inmode)
	endstr();
    columncount++, columncount %= column;
    charcount = 0;
    linecount = 0;
    if (columncount == 0) {
	putstr("grestore\n");
	putstr("showpage\n");
	putstr("restore\n");
	return (1);
    }
    startcolumn();
    return (0);
}

/*-==(spec:\\:":":\(:\):\[:\]::::psconv)==-----------------------------
    NAME
	cr - carriage return
 ---------------------------------------------------------------------*/
bool cr()
{
    linecount++;
    charcount = 0;
    if (linecount >= numrow)
	return (1);
    if (inmode)
	endstr();
    putstr("MOVE\n");
    return (0);
}

/*-==(spec:\\:":":\(:\):\[:\]::::psconv)==-----------------------------
    NAME
	reset
 ---------------------------------------------------------------------*/
void reset()
{
    putstr("nulldevice\n");
    putstr("initgraphics\n");
    putstr("initmatrix\n");
    putstr("grestoreall\n");
    putstr("clear\n");
}

/**********************************************************************
 *                                                                    *
 *  Begin/End-String Routines                                         *
 *                                                                    *
 **********************************************************************/
/*-==(spec:\\:":":\(:\):\[:\]::::psconv)==-----------------------------
    NAME
	beginstr
 ---------------------------------------------------------------------*/
void beginstr(mode, font)
    register enum chartype mode;
    register int font;
{
    if (currfont != font || currmode != mode)
	putstr(fonttable[(int) (currmode = mode)][currfont = font]);
    switch (mode) {
    case ASCII:
	putchar('(');
	break;
    case HANKAKU:
    case KANJI:
	putchar('<');
	break;
    default:
	break;
    }
    inmode = 1;
}

/*-==(spec:\\:":":\(:\):\[:\]::::psconv)==-----------------------------
    NAME
	endstr
 ---------------------------------------------------------------------*/
void endstr()
{
    switch (currmode) {
    case ASCII:
	putstr(") ASCII\n");
	break;
    case HANKAKU:
	putstr("> HANKAKU\n");
	break;
    case KANJI:
	putstr("> KANJI\n");
	break;
    default:
	break;
    }
    inmode = 0;
}

/**********************************************************************
 *                                                                    *
 *  Put-One-Character Routines                                        *
 *                                                                    *
 **********************************************************************/
/*-==(spec:\\:":":\(:\):\[:\]::::psconv)==-----------------------------
    NAME
	outchar
 ---------------------------------------------------------------------*/
void outchar(c, f)
    register int c, f;
{
    if (charcount >= width)
	if (cr() && ff())
	    startpage();
    if (inmode && (currmode != ASCII ||
	    (c == ' ' ? (currfont & U) != (f & U) : currfont != f)))
	endstr();
    if (!inmode)
	beginstr(ASCII, f);
    charcount++;
    if (c == '(' || c == ')' || c == '\\')
	putchar('\\');
    putchar(c);
}

/*-==(spec:\\:":":\(:\):\[:\]::::psconv)==-----------------------------
    NAME
	outhankana
 ---------------------------------------------------------------------*/
void outhankana(c, f)
    register int c, f;
{
    if (charcount >= width)
	if (cr() && ff())
	    startpage();
    if (inmode && (currmode != HANKAKU || currfont != f))
	endstr();
    if (!inmode)
	beginstr(HANKAKU, f);
    charcount++;
    printf("%04x ", (int) kanatable[c - 0xa0]);
}

/*-==(spec:\\:":":\(:\):\[:\]::::psconv)==-----------------------------
    NAME
	outkanji
 ---------------------------------------------------------------------*/
void outkanji(c, f)
    register int c, f;
{
    /*
     * Put at least one kanji character at the first column even if
     * width == 1.
     */
    if (charcount > 0 && charcount + 1 >= width)
	if (cr() && ff())
	    startpage();
    if (inmode && (currmode != KANJI || currfont != f))
	endstr();
    if (!inmode)
	beginstr(KANJI, f);
    charcount += 2;
    printf("%04x ", c & 0xffff);
}

/*-==(spec:\\:":":\(:\):\[:\]::::psconv)==-----------------------------
    NAME
	outnonprint
    DESCRIPTION
	"outnonprint()" prints non-printing characters in visible
	style; such as `[' overstruck by ``emdash'' for ESC (^[).
 ---------------------------------------------------------------------*/
void outnonprint(c, f)
    register int c, f;
{
    if (charcount >= width)
	if (cr() && ff())
	    startpage();
    if (inmode && (currmode != ASCII || currfont != f))
	endstr();
    if (!inmode)
	beginstr(ASCII, f);
    charcount++;
    if (c > 0x7f)
	putstr("\\277");	/* questiondown */
    else {
	c = c == 0x7f ? '?' : c | 0x40;
	if (c == '\\')
	    putchar('\\');
	putchar(c);
    }
    putstr(") NONPRINT\n");
    inmode = 0;
}

/**********************************************************************
 *                                                                    *
 *  Miscellaneous Routine                                             *
 *                                                                    *
 **********************************************************************/
/*-==(spec:\\:":":\(:\):\[:\]::::psconv)==-----------------------------
    NAME
	getlogname
 ---------------------------------------------------------------------*/
#include <pwd.h>

char *getlogname()
{
    char *user;
    char *getlogin();
    struct passwd *pwd;

    if ((user = getlogin()) == NULL)
	if ((pwd = getpwuid(getuid())) != NULL)
	    user = pwd->pw_name;
    return (user);
}

/**********************************************************************
 *                                                                    *
 *  Data Section                                                      *
 *                                                                    *
 **********************************************************************/
/*-==(spec9:\\:":":\(:\):\[:\]::::psconv)==----------------------------
    TYPE
	table
    NAME
	katakana
    DESCRIPTION
	JIS X0201 kana to JIS kanji (katakana) in EUC.
 ---------------------------------------------------------------------*/
unsigned short katakana[] = {
    0,      0xa1a3, 0xa1d6, 0xa1d7, 0xa1a2, 0xa1a6, 0xa5f2, 0xa5a1,
    /*                                              */
    0xa5a3, 0xa5a5, 0xa5a7, 0xa5a9, 0xa5e3, 0xa5e5, 0xa5e7, 0xa5c3,
    /*                                            */
    0xa1bc, 0xa5a2, 0xa5a4, 0xa5a6, 0xa5a8, 0xa5aa, 0xa5ab, 0xa5ad,
    /*                                            */
    0xa5af, 0xa5b1, 0xa5b3, 0xa5b5, 0xa5b7, 0xa5b9, 0xa5bb, 0xa5bd,
    /*                                            */
    0xa5bf, 0xa5c1, 0xa5c4, 0xa5c6, 0xa5c8, 0xa5ca, 0xa5cb, 0xa5cc,
    /*                                            */
    0xa5cd, 0xa5ce, 0xa5cf, 0xa5d2, 0xa5d5, 0xa5d8, 0xa5db, 0xa5de,
    /*                                            */
    0xa5df, 0xa5e0, 0xa5e1, 0xa5e2, 0xa5e4, 0xa5e6, 0xa5e8, 0xa5e9,
    /*                                            */
    0xa5ea, 0xa5eb, 0xa5ec, 0xa5ed, 0xa5ef, 0xa5f3, 0xa1ab, 0xa1ac,
    /*                                            */
};

/*-==(spec9:\\:":":\(:\):\[:\]::::psconv)==----------------------------
    TYPE
	table
    NAME
	hiragana
    DESCRIPTION
	JIS X0201 kana to JIS kanji (hiragana) in EUC.
 ---------------------------------------------------------------------*/
unsigned short hiragana[] = {
    0,      0xa1a3, 0xa1d6, 0xa1d7, 0xa1a2, 0xa1a6, 0xa4f2, 0xa4a1,
    /*                                            */
    0xa4a3, 0xa4a5, 0xa4a7, 0xa4a9, 0xa4e3, 0xa4e5, 0xa4e7, 0xa4c3,
    /*                                            */
    0xa1bc, 0xa4a2, 0xa4a4, 0xa4a6, 0xa4a8, 0xa4aa, 0xa4ab, 0xa4ad,
    /*                                            */
    0xa4af, 0xa4b1, 0xa4b3, 0xa4b5, 0xa4b7, 0xa4b9, 0xa4bb, 0xa4bd,
    /*                                            */
    0xa4bf, 0xa4c1, 0xa4c4, 0xa4c6, 0xa4c8, 0xa4ca, 0xa4cb, 0xa4cc,
    /*                                            */
    0xa4cd, 0xa4ce, 0xa4cf, 0xa4d2, 0xa4d5, 0xa4d8, 0xa4db, 0xa4de,
    /*                                            */
    0xa4df, 0xa4e0, 0xa4e1, 0xa4e2, 0xa4e4, 0xa4e6, 0xa4e8, 0xa4e9,
    /*                                            */
    0xa4ea, 0xa4eb, 0xa4ec, 0xa4ed, 0xa4ef, 0xa4f3, 0xa1ab, 0xa1ac,
    /*                                            */
};

/*-==(spec9:\\:":":\(:\):\[:\]::::psconv)==----------------------------
    TYPE
	string constant
    NAME
	header_str
 ---------------------------------------------------------------------*/
char *header_str[] = { "\
%!PS-Adobe-1.0\n\
%%Creator: psconv Ver 2.22 (PostScript filter by Y. Tonooka)\n\
%%For: ", "\n\
%%CreationDate: ", "\
%%Pages: (atend)\n\
%%DocumentFonts: Courier Courier-Oblique \
Courier-Bold Courier-BoldOblique \
Ryumin-Light-EUC-H GothicBBB-Medium-EUC-H\n\
%%EndComments\n\
%==(spec:\\\\:\":\":\\(:\\):\\[:\\]::::)==%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n\
%   TYPE\n\
%       page description\n\
%   NAME\n\
%       \"psconv\" output\n\
%   DESCRIPTION\n\
%       This PostScript program is fully conforming to the file\n\
%       structuring conventions version 1.0.\n\
%   NOTES\n\
%       For usage and details of \"psconv\", see `psconv.c' and manual\n\
%       page.\n\
%   LANGUAGE\n\
%       PostScript\n\
%   CREATOR\n\
%       \"psconv\" Ver 2.22 (by Yasuhiro Tonooka / Mar 28 1994)\n\
%       for ", "\n\
%   DATE\n\
%       ", "\
%==(endspec)==%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n\
\n" };

/*-==(spec9:\\:":":\(:\):\[:\]::::psconv)==----------------------------
    TYPE
	string constant
    NAME
	prolog_str - PostScript prologue string
 ---------------------------------------------------------------------*/
char prolog_str[] = "\
%\n\
% Procedures defined in small letters are \"local\" to Prolog block.\n\
%\n\
\n\
%\n\
% Fonts for ASCII characters:\n\
%\n\
/ascii-r /Courier findfont point scalefont def\n\
/ascii-i /Courier-Oblique findfont point scalefont def\n\
/ascii-b /Courier-Bold findfont point scalefont def\n\
/ascii-c /Courier-BoldOblique findfont point scalefont def\n\
/R { ascii-r setfont } def\n\
/I { ascii-i setfont } def\n\
/B { ascii-b setfont } def\n\
/C { ascii-c setfont } def\n\
\n\
%\n\
% Kanji font dictionaries are made only when needed.\n\
%\n\
/min { /Ryumin-Light-EUC-H findfont } def\n\
/goth { /GothicBBB-Medium-EUC-H findfont } def\n\
/kanji-r { [ kanji-width point mul 0 0 kanji-height point mul 0 0 ] makefont } def\n\
/kanji-i { [ kanji-width point mul 0 .23 kanji-height point mul mul kanji-height point mul 0 0 ]\n\
\tmakefont } def\n\
/hankaku-r { [ hankaku-width point mul 0 0 hankaku-height point mul 0 0 ] makefont } def\n\
/hankaku-i { [ hankaku-width point mul 0 .23 hankaku-height point mul mul hankaku-height point mul 0 0 ] makefont } def\n\
/kr null def\n\
/ki null def\n\
/kb null def\n\
/kc null def\n\
/hr null def\n\
/hi null def\n\
/hb null def\n\
/hc null def\n\
\n\
/KR { kr type (nulltype) eq { /kr min kanji-r def } if kr setfont } def\n\
/KI { ki type (nulltype) eq { /ki min kanji-i def } if ki setfont } def\n\
/KB { kb type (nulltype) eq { /kb goth kanji-r def } if kb setfont } def\n\
/KC { kc type (nulltype) eq { /kc goth kanji-i def } if kc setfont } def\n\
/HR { hr type (nulltype) eq { /hr min hankaku-r def } if hr setfont } def\n\
/HI { hi type (nulltype) eq { /hi min hankaku-i def } if hi setfont } def\n\
/HB { hb type (nulltype) eq { /hb goth hankaku-r def } if hb setfont } def\n\
/HC { hc type (nulltype) eq { /hc goth hankaku-i def } if hc setfont } def\n\
\n\
%\n\
% \"show\" procedures:\n\
%\n\
/ASCII {\n\
\tdoul { { pop pop underline } exch kshow underline }\n\
\t{ show } ifelse } def\n\
/KANJI {\n\
\ten <a1a1> stringwidth pop 2 div sub\n\
\t0 2 3 index length 1 sub {\n\
\t\t1 index 0 rmoveto\n\
\t\t2 index exch 2 getinterval show\n\
\t\tdup 0 rmoveto\n\
\t\tdoul { backspace backspace\n\
\t\t\tcurrentfont R (__) show setfont } if\n\
\t} for\n\
\tpop pop } def\n\
/HANKAKU {\n\
\ten <a1a1> stringwidth pop sub 2 div\n\
\t0 2 3 index length 1 sub {\n\
\t\t1 index 0 rmoveto\n\
\t\t2 index exch 2 getinterval show\n\
\t\tdup 0 rmoveto\n\
\t\tdoul { underline } if\n\
\t} for\n\
\tpop pop } def\n\
/NONPRINT { % Show an ASCII character then overstrike with emdash:\n\
\tASCII backspace (\\320) show } def\n\
\n\
%\n\
% Other global procedures:\n\
%\n\
/MOVE { left-margin y-pos moveto /y-pos y-pos linefeed sub def } def\n\
\n\
/U { /doul true def } def\t% underlining begin\n\
/N { /doul false def } def\t% underlining end\n\
\n\
/PAGE-SETUP {\n\
\tinit-coordinate\n\
\t/left-margin column-interval en mul neg def\n\
\tdraw-dash } def\n\
/COLUMN-SETUP {\n\
\t/left-margin left-margin column-interval en mul add def\n\
\t/y-pos 0 def } def\n\
\n\
%\n\
% Local procedures:\n\
%\n\
/backspace { en neg 0.0 rmoveto } def\n\
/underline { backspace currentfont R (_) show setfont } def\n\
\n\
/init-coordinate {\t\t%initialize coordinate\n\
\tbase-r rotate base-x base-y translate\n\
\tbase-scale-x base-scale-y scale } def\n\
\n\
/draw-dash {\t\t\t% draw separator (dash)\n\
\t0.22 setlinewidth\n\
\t1 setlinecap\n\
\t[ 1.2 2.4 ] 0 setdash\n\
\tdash-init\n\
\t2 1 dash-count {\n\
\t\tpop\n\
\t\tdup en mul 0 moveto\n\
\t\t0 dash-len neg rlineto\n\
\t\tstroke\n\
\t\tcolumn-interval add\n\
\t} for pop } def\n\
\n\
%\n\
% Get the width of en character.\n\
%\n\
gsave\n\
init-coordinate\n\
R /en ( ) stringwidth pop def\n\
grestore\n\
\n\
%%EndProlog\n";
